/* Ne3sML.c 
   This calls mathematica to calculate the likelihood function
  
   MS Windows/MS VC++ 6:
     cl -O2 -W3 Ne3sML.c minsub2.c -link ml32i2m.lib
     Ne3sML -linkmode launch

   Windows/Borland C++/Mathematica v4:
     bcc32 -O2 Ne3sML.c minsub2.c -link ml32i2b.lib
     JC3sBPP -linkmode launch

   UNIX: Ne3s -linkname "math -mathlink"

   MAC: 
     JC3sBPP -linkmode launch

*/

#include "mathlink.h"
#include "minsub.h"
#include <time.h>

extern double Small_Diff;
extern int NFunCall;

void error2 (char * message);
void starttime (void);
char* printtime (char timestr[]);
double lfun(double x[], int np);
int ming2 (FILE *fout, double *f, double (*fun)(double x[], int n),
    int (*dfun)(double x[], double *f, double dx[], int n),
    double x[], double xb[][2], double space[], double e, int n);

static void error1(MLINK link);
static void init_and_openlink(int argc, char* argv[]);

MLENV env = (MLENV)0;
MLINK link = (MLINK)0;

int nloci, *nij;
extern int noisy, NFunCall;

int main (int argc, char* argv[])
{
   FILE *fin, *frub;
   int np=4, i, nsites1=0,nsites2=0, iterate=1;
   char datafile[96]="Ne3sML.dat", mathpackfile[96]="Ne3slnL.m", timestr[96];
   double space[1000], *var=space+np, lnL=-1, e=1e-6, xb[4][2];
   double a=2.2, b=3.3, sumab=-1;
   double x[4]={0.3057, 0.0990, 0.1089, 0.5194};  /* lnL = -3099.41 */

   if(argc!=3 && argc!=4) {
       puts("\n\tUsage:  Ne3slnL -linkmode launch <mathpackage>\n");
       exit(-1);
   }
   if(argc==4) strcpy(mathpackfile, argv[3]);
   printf("\nlnL calculated using Mathematica package %s\n", mathpackfile);

   if((fin=(FILE*)fopen(datafile,"r"))==NULL) error2("data file open error");
   if((frub=(FILE*)fopen("rub","w"))==NULL) error2("rub open error");
   fscanf(fin, "%d", &nloci);
   if((nij=(int*)malloc(nloci*5*sizeof(int)))==NULL) error2("oom");
   for(i=0,nsites1=0; i<nloci*5; i++) {
      fscanf(fin, "%d", &nij[i]);  nsites1+=nij[i];
   }
   fclose(fin);
   printf("Please locate and select MathKernal\n");

   init_and_openlink(argc, argv);

   MLPutFunction(link, "Needs", 2L);
      MLPutString(link, "Ne3slnL`");
      MLPutString(link, mathpackfile);
   MLEndPacket(link);
   while(MLNextPacket(link) != RETURNPKT) {
      MLNewPacket(link);
      if (MLError(link)) error1(link);
   }
   MLNewPacket(link);  /* abandon the RETURNPKT */

   MLPutFunction(link, "EvaluatePacket", 1L);
      MLPutFunction(link, "AddTwo", 2L);
         MLPutReal(link, a);
         MLPutReal(link, b);
   MLEndPacket(link);
   while(MLNextPacket(link) != RETURNPKT) {
      MLNewPacket(link);
      if (MLError(link)) error1(link);
   }
   MLGetReal(link, &sumab);
   printf("%.4f + %.4f = %.4f\n", a,b,sumab);

   MLPutFunction(link, "EvaluatePacket", 1L);
      MLPutFunction(link, "SumNij", 1L);
         MLPutIntegerList(link, nij, nloci*5);
   MLEndPacket(link);
   while(MLNextPacket(link) != RETURNPKT) {
      MLNewPacket(link);
      if (MLError(link)) error1(link);
   }
   MLGetInteger(link, &nsites2);
   printf("%d loci, %d = %d sites\n\n", nloci, nsites1, nsites2);

   noisy=3;
   printf("Iterate? (1 for Yes; 0 for No, which means lnL and SE at the values given) ");
   scanf("%d", &iterate);
   for(i=0; i<np; i++) { xb[i][0]=0.0001; xb[i][1]=99; }
   for(i=0; i<np; i++) x[i]=.5;
   printf("Input 4 initial values for theta_0 theta_1 gamma_0 gamma_1 (from .05 to 2):\n");
   printf("Note that all four parameters are multiplied by 100 in the program\n");
   for(i=0; i<np; i++) scanf("%lf", &x[i]);
   printf("Input a small step length about 1e-3 to 1e-8 (current %.6f):\n", Small_Diff);
   scanf("%lf", &Small_Diff);
   printf("Current parameter values: ");

   starttime();
   for(i=0; i<np; i++) printf(" %9.5f", x[i]);  
   printf("\n");
   lnL=lfun(x, np);
   printf("lnL0 = %12.6f\n", -lnL);
   printf("\nTime used: %s (iteration takes about 100 times as long)\n", printtime(timestr));

   NFunCall=0;
   if(iterate) 
      i=ming2(frub, &lnL, lfun, NULL, x, xb, space, e, np);

   printf ("Out...\nlnL  = %12.6f  (Time used: %s)\nMLEs: ", -lnL, printtime(timestr));
   for(i=0; i<np; i++) printf(" %9.5f", x[i]);  printf("\n");

   puts("Calculating approximate SEs and variance matrix.  Ctrl-C to break.");
   Hessian (np,x,lnL,space,var,lfun,var+np*np);
   matinv(var,np,np,var+np*np);
   printf("SEs:  ");
   for(i=0;i<np;i++) printf(" %9.5f",(var[i*np+i]>0.?sqrt(var[i*np+i]):-1));
   printf("\n");
   matout2(stdout, var, np, np, 19, 12);

   printf ("#lnL function calls: %d\n", NFunCall);

   MLPutFunction(link, "Exit", 0);
   free(nij);
   printf("\nTime used: %s\n", printtime(timestr));
   return 0;
}



static time_t time_start;

void starttime (void)
{
   time_start=time(NULL);
}

char* printtime (char timestr[])
{
/* print time elapsed since last call to starttime()
*/
   time_t t;
   int h, m, s;

   t=time(NULL)-time_start;
   h=t/3600; m=(t%3600)/60; s=t-(t/60)*60;
   if(h)  sprintf(timestr,"%d:%02d:%02d", h,m,s);
   else   sprintf(timestr,"%2d:%02d", m,s);
   return(timestr);
}

double lfun (double x[], int np)
{
   double lnL=0;

   NFunCall++;

   MLPutFunction(link, "EvaluatePacket", 1L);
      MLPutFunction(link, "lfunNe3s", 5L);
         MLPutReal(link, x[0]);
         MLPutReal(link, x[1]);
         MLPutReal(link, x[2]);
         MLPutReal(link, x[3]);
         MLPutIntegerList(link, nij, (long)(nloci*5L));
   MLEndPacket(link);
   while(MLNextPacket(link) != RETURNPKT) {
      MLNewPacket(link);
      if (MLError(link)) error1(link);
   }

   MLGetReal(link, &lnL);
   return (lnL);
}


static void error1(MLINK link)
{
   if(MLError(link)){
      fprintf(stderr, "Error detected by MathLink: %s.\n",
         MLErrorMessage(link));
   }else{
      fprintf(stderr, "Error detected by this program.\n");
   }
   exit(3);
}

static void deinit(void)
{
   if(env) MLDeinitialize(env);
}

static void closelink(void)
{
   if(link) MLClose(link);
}

static void init_and_openlink(int argc, char* argv[])
{
   long err;

   env =  MLInitialize((MLParametersPointer)0);
   if(env == (MLENV)0) exit(1);
   atexit(deinit);

   link = MLOpenArgv(env, argv, argv + argc, &err);
   if(link == (MLINK)0) exit(2);
   atexit(closelink);
}
