(* :Title:   Ne3slnL          *)
BeginPackage["Ne3slnL`"]

AddTwo::usage = "Usage: AddTwo[Real x, Real y]";
SumNij::usage = "Usage: SumNij[RealList x]";
lfunNe3s::usage="lfunNe3s[theta0, theta1, gamma0, gamma1, nij] calculates the \
lnL for Ne3s.";


Begin["`Private`"]

AddTwo[x_Real, y_Real] := x + y;

SumNij[nij_List]:=
Module[{i=0,n=Length[nij],sum=0},
   While[(i=i+1)<=n,sum+=nij[[i]]];
   sum
]

IntegrandCaseI = Compile[
   {{x0,_Real}, {x1,_Real}, {theta0,_Real}, {theta1,_Real}, {gamma0,_Real}, {gamma1,_Real},
    {ScaleFactor,_Real}, 
    {n0,_Integer}, {n1,_Integer}, {n2,_Integer}, {n3,_Integer}, {n4,_Integer}},

   Module[{p0,p1,p2,p4, b0,b1, e1,e2,e3, lnP, Small=-200.0},
      b0=gamma0+theta0*x0/2.-theta1*x1/2.;   b1=gamma1+theta1*x1/2.;
      e1= Exp[-4./3*b1 ];  e2 = Exp[-8./3*(b0+b1)];
      e3 =e1* e2;  e1 = e1*e1;
      p0=(1 + 3 e1 + 6 e2 + 6 e3)/16;
      p1=(3 + 9 e1 - 6 e2 - 6 e3)/16;
      p2=(3 - 3 e1 + 6 e2 - 6 e3)/16;
      p4=(6 - 6 e1 -12 e2 +12 e3)/16;
      lnP = n0*Log[p0]+n1*Log[p1]+(n2+n3)*Log[p2]+n4*Log[p4] - (x0+x1)- ScaleFactor;
      If[lnP<Small, 0.,Exp[lnP]]
   ]
]

IntegrandCaseII3trees = Compile[
   {{x0,_Real}, {x1,_Real}, {theta0,_Real}, {theta1,_Real}, {gamma0,_Real}, {gamma1,_Real},
    {ScaleFactor,_Real}, 
    {n0,_Integer}, {n1,_Integer}, {n2,_Integer}, {n3,_Integer}, {n4,_Integer}},

   Module[{p0,p1,p2,p4, b0,b1, e1,e2,e3, p12, lnP, Small=-200.0},
      b0=theta0*x0/2.;  b1=gamma0+gamma1+theta0*x1/2.;
      e1 = Exp[-4./3*b1];  e2 = Exp[-8./3*(b0+b1)];
      e3 = e1* e2;  e1 = e1*e1;
      p0=(1 + 3 e1 + 6 e2 + 6 e3)/16;
      p1=(3 + 9 e1 - 6 e2 - 6 e3)/16;
      p2=(3 - 3 e1 + 6 e2 - 6 e3)/16;
      p4=(6 - 6 e1 -12 e2 +12 e3)/16;
      p12=p1/p2;
      lnP = n0*Log[p0] + (n1+n2+n3)*Log[p2]+n4*Log[p4] + Log[p12^n1 + p12^n2 + p12^n3]
          - (x0+3*x1) - ScaleFactor;
      If[lnP<Small, 0., Exp[lnP]]
   ]
]

lfunNe3s[theta0_Real,theta1_Real, gamma0_Real, gamma1_Real, nij_List] :=
     Module[ {nloci, locus, alnL, fh, PI, PII3trees, n0,n1,n2,n3,n4,n,  
           th0,th1,ga0,ga1, ratelocus=.0, ScaleFactor, Small=-200.0},
     nloci=Length[nij]/5; locus=0; alnL=0.;
     While[(locus=locus+1)<=nloci,  
        n0=nij[[(locus-1)*5+1]];  n1=nij[[(locus-1)*5+2]]; n2=nij[[(locus-1)*5+3]];
        n3=nij[[(locus-1)*5+4]];  n4=nij[[(locus-1)*5+5]];
        n=(n0+n1+n2+n3+n4)*1.0;
        (*ratelocus=rates[[locus]];*)
        ratelocus=1.0;
        th0=theta0/100.*ratelocus;  th1=theta1/100.*ratelocus;
        ga0=gamma0/100.*ratelocus;  ga1=gamma1/100.*ratelocus;

        ScaleFactor = Log[n0/n]*n0;
        ScaleFactor = If[n0<n, ScaleFactor+(n-n0)*Log[1.-n0/n], ScaleFactor];

        PI = NIntegrate[IntegrandCaseI[x0,x1,th0, th1, ga0, ga1, ScaleFactor, n0,n1,n2,n3,n4],
                        {x0, 0., Infinity}, {x1,0.,2.*ga0/th1}];
        PII3trees = NIntegrate[ 
                IntegrandCaseII3trees[x0,x1,th0, th1, ga0, ga1, ScaleFactor, n0,n1,n2,n3,n4],
                 {x0, 0., Infinity}, {x1,0.,Infinity}] * Exp[-2.*ga0/th1];
        PI = PI+PII3trees;
        (* If[PI< Exp[Small],  Print["small p", PI], fh=-1]; *)
        (* If[PI>10^300,  Print["large p", PI], fh=-1];  *)
        fh = Log[PI ]+ ScaleFactor;
        alnL-=fh;
        (* Print[{locus,nloci, n0,n1,n2,n3,n4, ScaleFactor}]  *)
     ] ;
     alnL
]

End[]
EndPackage[]
